/**
 * Copyright (C) 2006-2018 INRIA and contributors
 * Spoon - http://spoon.gforge.inria.fr/
 *
 * This software is governed by the CeCILL-C License under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/or redistribute the software under the terms of the CeCILL-C license as
 * circulated by CEA, CNRS and INRIA at http://www.cecill.info.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the CeCILL-C License for more details.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C license and that you accept its terms.
 */
package spoon.test.method_overriding;

import spoon.SpoonModelBuilder;
import spoon.reflect.factory.Factory;
import spoon.test.method_overriding.testclasses2.ObjectInterface;
import spoon.compiler.SpoonResourceHelper;
import spoon.reflect.visitor.filter.TypeFilter;
import spoon.reflect.declaration.CtType;
import spoon.testing.utils.ModelUtils;
import spoon.Launcher;
import spoon.reflect.declaration.CtMethod;
import org.junit.jupiter.api.Test;

import java.util.HashMap;
import java.util.ArrayList;
import java.util.function.BiFunction;
import java.util.Comparator;
import java.util.List;
import java.io.File;
import java.util.Map;

import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.assertFalse;

public class MethodOverriddingTest {

	@Test
	public void testShadowInterfaceMethodsCanOverrideObjectMethods() {
		//contract: Interface (made by reflection) method equals overrides Object#equals
		Factory f = new Launcher().getFactory();
		CtType<?> iface = f.Interface().get(Comparator.class);
		assertTrue(iface.isShadow());
		CtMethod<?> comparatorEquals = iface.getMethodsByName("equals").get(0);
		
		CtType<?> object = f.Class().get(Object.class);
		assertTrue(object.isShadow());
		CtMethod<?> objectEquals = object.getMethodsByName("equals").get(0);
		
		assertTrue(comparatorEquals.isOverriding(objectEquals));
	}

	@Test
	public void testInterfaceMethodsCanOverrideObjectMethods() throws Exception {
		//contract: Interface (made from sources) method equals overrides Object#equals
		Launcher launcher = new Launcher();
		Factory f = launcher.getFactory();
		SpoonModelBuilder comp = launcher.createCompiler();
		comp.addInputSources(SpoonResourceHelper.resources("./src/test/java/spoon/test/method_overriding/testclasses2/ObjectInterface.java"));
		comp.build();
		CtType<?> iface = f.Interface().get(ObjectInterface.class);
		assertFalse(iface.isShadow());
		CtMethod<?> ifaceEquals = iface.getMethodsByName("equals").get(0);
		
		CtType<?> object = iface.getFactory().Class().get(Object.class);
		assertTrue(object.isShadow());
		CtMethod<?> objectEquals = object.getMethodsByName("equals").get(0);
		
		assertTrue(ifaceEquals.isOverriding(objectEquals));
	}

	@Test
	public void testMethodOverride() {
		checkMethodOverride((m1, m2)->m1.isOverriding(m2));
	}

	@Test
	public void testMethodOverrideByReference() {
		checkMethodOverride((m1, m2)->m1.getReference().isOverriding(m2.getReference()));
	}

	private void checkMethodOverride(BiFunction<CtMethod<?>, CtMethod<?>, Boolean> isOverriding) {
		Factory factory = ModelUtils.build(new File("src/test/java/spoon/test/method_overriding/testclasses").listFiles());
		Map<String, List<CtMethod>> methodsByName = new HashMap<>();
		factory.getModel().filterChildren(new TypeFilter<>(CtMethod.class)).forEach((CtMethod m)->{
			List<CtMethod> methods = methodsByName.get(m.getSimpleName());
			if(methods==null) {
				methods = new ArrayList<>();
				methodsByName.put(m.getSimpleName(), methods);
			}
			methods.add(m);
		});
		assertFalse(methodsByName.isEmpty());
		for (Map.Entry<String, List<CtMethod>> e : methodsByName.entrySet()) {
			combine(e.getValue(), 0, isOverriding);
		}
	}

	private void combine(List<CtMethod> value, int start, BiFunction<CtMethod<?>, CtMethod<?>, Boolean> isOverriding) {
		CtMethod m1 = value.get(start);
		if(start+1 < value.size()) {
			for (CtMethod m2 : value.subList(start+1, value.size())) {
				if(m1.getDeclaringType().isSubtypeOf(m2.getDeclaringType().getReference())) {
					checkOverride(m1, m2, isOverriding);
				} else if(m2.getDeclaringType().isSubtypeOf(m1.getDeclaringType().getReference())) {
					checkOverride(m2, m1, isOverriding);
				} else {
					checkNotOverride(m1, m2, isOverriding);
				}
			}
			combine(value, start+1, isOverriding);
		}
	}

	private void checkOverride(CtMethod m1, CtMethod m2, BiFunction<CtMethod<?>, CtMethod<?>, Boolean> isOverriding) {
		assertTrue(isOverriding.apply(m1, m2), descr(m1) + " overriding " + descr(m2));
		assertFalse(isOverriding.apply(m2, m1), descr(m2) + " NOT overriding " + descr(m1));
	}

	private void checkNotOverride(CtMethod m1, CtMethod m2, BiFunction<CtMethod<?>, CtMethod<?>, Boolean> isOverriding) {
		assertFalse(isOverriding.apply(m1, m2), descr(m1) + " NOT overriding " + descr(m2));
		assertFalse(isOverriding.apply(m2, m1), descr(m2) + " NOT overriding " + descr(m1));
	}

	private String descr(CtMethod m) {
		return m.getDeclaringType().getSimpleName()+"#"+m.getSimpleName();
	}
}
